/*
 * Copyright 2009 OW2 Chameleon
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *   http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.ow2.chameleon.naming.jndi;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Dictionary;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.List;
import java.util.Properties;

import javax.naming.Binding;
import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.Name;
import javax.naming.NameClassPair;
import javax.naming.NameParser;
import javax.naming.NamingEnumeration;
import javax.naming.NamingException;

import org.apache.felix.ipojo.annotations.Component;
import org.apache.felix.ipojo.annotations.Invalidate;
import org.apache.felix.ipojo.annotations.Property;
import org.apache.felix.ipojo.annotations.Provides;
import org.apache.felix.ipojo.annotations.ServiceProperty;
import org.osgi.framework.Bundle;
import org.osgi.framework.BundleContext;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

/**
 * Component Implementing the naming service ({@link Context}) based on jndi.
 * The instance is configured either using a properties or the path on the
 * <tt>jndi.properties</tt> file. So, instance configuration must contains
 * either
 * <ul>
 * <li><tt>jndi.file</tt> : path to the <tt>jndi.properties</tt> file</li>
 * <li><tt>jndi.configuration</tt> : dictionary containing the jndi
 * configuration
 * </ul>
 * One of them is mandatory to get the naming service exposed.
 *
 * The configuration can also define the <tt>jndi.id</tt> property to identify
 * the jndi service used. If this property is set the service exposes the <tt>
 * jndi.id</tt> property, otherwise this service property is ignored.
 *
 * This implementation uses a bridge classloader in order to load the naming
 * service provider class. The bridge classloader use an heuristic policy to
 * avoid checking all bundles every time. The bridge can be configured to
 * exclude some bundles (known to be error-prone) from the bridge. So the
 * <tt>excluded.bundles</tt> instance property allow excluding bundles from the
 * bridge. It receives a comma-separated list of bundle symbolic name.
 *
 * For example, the following snippet is a valid configuration for this
 * implementation:
 * <pre>
 * excluded.bundles=osgi.cmpn
 * jndi.file=conf/jndi.properties
 * jndi.id=localhost
 * </pre>
 * It excludes the OSGi Compendium from the bridge (because it does a
 * dynamic import on everything). It also configures the jndi provider using a
 * <tt>jndi.properties</tt> from the <tt>conf</tt> folder.
 *
 * This bundles <b>does not contain</b> any naming service implementation. They
 * must be provided by others bundles, exporting the correct set of classes.
 *
 * Finally, the created instance exposes the {@link Context} interface as a service
 * and also publishes the <tt>java.naming.provider.url</tt> service property.
 */
@Component(immediate = true)
@Provides
public class JNDINamingService implements Context {

    private static final Logger LOG = LoggerFactory
            .getLogger(JNDINamingService.class);

    private final BundleContext m_bundleContext;

    private Context m_delegate;

    private Properties m_configuration;

    @Property(name = "excluded.bundles")
    private String m_excludedBundleList;

    @ServiceProperty(name = "java.naming.provider.url")
    String m_jndiURL;

    @ServiceProperty(name = "jndi.id")
    String m_jndiId;

    public JNDINamingService(BundleContext bc) {
        m_bundleContext = bc;
    }

    @Property(name = "jndi.file")
    public void setFile(String fileName) throws FileNotFoundException,
            IOException, NamingException {
        LOG.info("Receving jndi.file configuration " + fileName);
        File file = new File(fileName);
        m_configuration = new Properties();
        if (file.exists()) {
            LOG.info("Loading jndi configuration from "
                    + file.getAbsolutePath());
            m_configuration.load(new FileInputStream(file));
            LOG.info("jndi configuration loaded from " + file.getAbsolutePath());
        } else {
            LOG.info("Loading jndi configuration from "
                    + m_bundleContext.getBundle().getSymbolicName());
            URL url = m_bundleContext.getBundle().getResource(fileName);
            if (url == null) {
                LOG.info("Cannot load jndi configuration from " + fileName
                        + " : the resource is not in the bundle");
                throw new FileNotFoundException(
                        "Cannot load jndi configuration from " + fileName
                                + " : the resource is not in the bundle");
            } else {
                LOG.info("Loading jndi configuration loaded from "
                        + url.toExternalForm());
                m_configuration.load(url.openStream());
                LOG.info("jndi configuration loaded from "
                        + url.toExternalForm());
            }
        }
        configure();
    }

    @Property(name = "jndi.configuration")
    public void setConfiguration(Dictionary<String, String> props)
            throws NamingException {
        m_configuration = new Properties();
        LOG.info("Loading jndi configuration loaded from configuration");
        Enumeration<String> e = props.elements();
        while (e.hasMoreElements()) {
            String key = e.nextElement();
            m_configuration.put(key, props.get(key));
        }
        LOG.info("jndi configuration loaded from configuration");
        configure();
    }

    private void configure() throws NamingException {
        LOG.info("Initializing naming service delegate");
        ClassLoader original = Thread.currentThread().getContextClassLoader();
        try {
            // Switch classloader
            Thread.currentThread().setContextClassLoader(
                    new BridgeClassLoader());
            m_delegate = new InitialContext(m_configuration);
            m_jndiURL = m_configuration.getProperty("java.naming.provider.url");

            LOG.info("Naming service delegate initialized : " + m_delegate
                    + " with url " + m_jndiURL + " (id : " + m_jndiId + " )");
        } finally {
            // Back to the original
            Thread.currentThread().setContextClassLoader(original);
        }
    }

    @Invalidate
    public void stop() throws NamingException {
        if (m_delegate != null) {
            m_delegate.close();
            m_delegate = null;
        }
    }

    public Object lookup(Name name) throws NamingException {
        LOG.info("Lookup " + name);
        ClassLoader original = Thread.currentThread().getContextClassLoader();
        try {
            // Switch classloader
            Thread.currentThread().setContextClassLoader(
                    new BridgeClassLoader());
            return m_delegate.lookup(name);
        } finally {
            // Back to the original
            Thread.currentThread().setContextClassLoader(original);
        }
    }

    public Object lookup(String name) throws NamingException {
        LOG.info("Lookup " + name);
        ClassLoader original = Thread.currentThread().getContextClassLoader();
        try {
            // Switch classloader
            Thread.currentThread().setContextClassLoader(
                    new BridgeClassLoader());
            return m_delegate.lookup(name);
        } finally {
            // Back to the original
            Thread.currentThread().setContextClassLoader(original);
        }
    }

    public void bind(Name name, Object obj) throws NamingException {
        m_delegate.bind(name, obj);
    }

    public void bind(String name, Object obj) throws NamingException {
        m_delegate.bind(name, obj);
    }

    public void rebind(Name name, Object obj) throws NamingException {
        m_delegate.rebind(name, obj);
    }

    public void rebind(String name, Object obj) throws NamingException {
        m_delegate.rebind(name, obj);
    }

    public void unbind(Name name) throws NamingException {
        m_delegate.unbind(name);
    }

    public void unbind(String name) throws NamingException {
        m_delegate.unbind(name);
    }

    public void rename(Name oldName, Name newName) throws NamingException {
        m_delegate.rename(oldName, newName);
    }

    public void rename(String oldName, String newName) throws NamingException {
        m_delegate.rename(oldName, newName);
    }

    public NamingEnumeration<NameClassPair> list(Name name)
            throws NamingException {
        return m_delegate.list(name);
    }

    public NamingEnumeration<NameClassPair> list(String name)
            throws NamingException {
        return m_delegate.list(name);
    }

    public NamingEnumeration<Binding> listBindings(Name name)
            throws NamingException {
        return m_delegate.listBindings(name);
    }

    public NamingEnumeration<Binding> listBindings(String name)
            throws NamingException {
        return m_delegate.listBindings(name);
    }

    public void destroySubcontext(Name name) throws NamingException {
        m_delegate.destroySubcontext(name);
    }

    public void destroySubcontext(String name) throws NamingException {
        m_delegate.destroySubcontext(name);
    }

    public Context createSubcontext(Name name) throws NamingException {
        return m_delegate.createSubcontext(name);
    }

    public Context createSubcontext(String name) throws NamingException {
        return m_delegate.createSubcontext(name);
    }

    public Object lookupLink(Name name) throws NamingException {
        LOG.info("Lookup link " + name);
        ClassLoader original = Thread.currentThread().getContextClassLoader();
        try {
            // Switch classloader
            Thread.currentThread().setContextClassLoader(
                    new BridgeClassLoader());
            return m_delegate.lookupLink(name);
        } finally {
            // Back to the original
            Thread.currentThread().setContextClassLoader(original);
        }
    }

    public Object lookupLink(String name) throws NamingException {
        LOG.info("Lookup link " + name);
        ClassLoader original = Thread.currentThread().getContextClassLoader();
        try {
            // Switch classloader
            Thread.currentThread().setContextClassLoader(
                    new BridgeClassLoader());
            return m_delegate.lookupLink(name);
        } finally {
            // Back to the original
            Thread.currentThread().setContextClassLoader(original);
        }
    }

    public NameParser getNameParser(Name name) throws NamingException {
        return m_delegate.getNameParser(name);
    }

    public NameParser getNameParser(String name) throws NamingException {
        return m_delegate.getNameParser(name);
    }

    public Name composeName(Name name, Name prefix) throws NamingException {
        return m_delegate.composeName(name, prefix);
    }

    public String composeName(String name, String prefix)
            throws NamingException {
        return m_delegate.composeName(name, prefix);
    }

    public Object addToEnvironment(String propName, Object propVal)
            throws NamingException {
        return m_delegate.addToEnvironment(propName, propVal);
    }

    public Object removeFromEnvironment(String propName) throws NamingException {
        return m_delegate.removeFromEnvironment(propName);
    }

    public Hashtable<?, ?> getEnvironment() throws NamingException { // NOSONAR
        return m_delegate.getEnvironment();
    }

    public void close() throws NamingException {
        throw new UnsupportedOperationException(
                "you shouldn't close this Context on your own");
    }

    public String getNameInNamespace() throws NamingException {
        return m_delegate.getNameInNamespace();
    }

    /**
     * Bridge classloader delegating classloading to:
     * <ol>
     * <li>the current bundle</li>
     * <li>the latest used bundle if still here</li>
     * <li>any (non-exlucded) bundle from the installed bundle set able to load the class.
     * If a bundle is able to load the class, the bundle is kept for the next
     * query (heuristic)</li>
     * </ol>
     */
    private class BridgeClassLoader extends ClassLoader {

        private long m_last = -1;
        private List<String> m_excludedBundles = new ArrayList<String>();

        public BridgeClassLoader() {
            setExcludedBundleList();
        }

        @Override
        public Class<?> loadClass(String name) throws ClassNotFoundException {

            try {
                LOG.debug("Trying to load class " + name + " from "
                        + m_bundleContext.getBundle().getSymbolicName());
                return m_bundleContext.getBundle().loadClass(name);
            } catch (ClassNotFoundException e) {
                if (m_last != -1 && m_bundleContext.getBundle(m_last) != null) {
                    try {
                        LOG.debug("Trying to load class "
                                + name
                                + " from latest bundle: "
                                + m_bundleContext.getBundle(m_last)
                                        .getSymbolicName());
                        return m_bundleContext.getBundle(m_last)
                                .loadClass(name);
                    } catch (ClassNotFoundException e2) {
                        // Will try next.
                    }
                }

                // Try all bundles.
                Bundle[] bundles = m_bundleContext.getBundles();
                for (Bundle bundle : bundles) {
                    try {
                        if (!m_excludedBundles.contains(bundle
                                .getSymbolicName())) {
                            Class<?> clazz = bundle.loadClass(name);
                            // Success.
                            LOG.debug("Class " + name + " loaded from "
                                    + bundle.getSymbolicName());
                            // Set last
                            m_last = bundle.getBundleId();
                            return clazz;
                        } else {
                            LOG.debug("Bundle excluded from bridge: "
                                    + bundle.getSymbolicName());
                        }
                    } catch (ClassNotFoundException e2) {
                        // Ignore.
                    }
                }
            }

            return super.loadClass(name);
        }

        private void setExcludedBundleList() {
            if (m_excludedBundleList == null || m_excludedBundleList.length() == 0) {
                return;
            }

            m_excludedBundles.clear();
            String[] bundles = m_excludedBundleList.trim().split(",");
            for (String bundle : bundles) {
                m_excludedBundles.add(bundle.trim());
            }
            LOG.info("Excluding bundles: " + m_excludedBundles);
        }

    }

}
